#include <stdlib.h>
#include <string.h>

#include "huffman.h"

class CHuffmanNode
{
public:
	CHuffmanNode() {	memset(this, 0, sizeof(CHuffmanNode));	}

	int nFrequency;	// must be first for byte align
	unsigned char byAscii;
	unsigned int dwCode;
	int nCodeLength;
	CHuffmanNode *pParent, *pLeft, *pRight;
};

int frequencyCompare(const void *elem1, const void *elem2 )
{
	CHuffmanNode *pNodes[2] = {	(CHuffmanNode*)elem1, (CHuffmanNode*)elem2 };
	if(pNodes[0]->nFrequency == pNodes[1]->nFrequency)
		return 0;
	return pNodes[0]->nFrequency < pNodes[1]->nFrequency ? 1 : -1;
}

int asciiCompare(const void *elem1, const void *elem2 )
{
	return ((CHuffmanNode*)elem1)->byAscii > ((CHuffmanNode*)elem2)->byAscii ? 1 : -1;
}

inline CHuffmanNode* PopNode(CHuffmanNode *pNodes[], int nIndex, bool bRight)
{
	register CHuffmanNode* pNode = pNodes[nIndex];
	pNode->dwCode = bRight;
	pNode->nCodeLength = 1;
	return pNode;
}

inline void SetNodeCode(CHuffmanNode* pNode)
{
	register CHuffmanNode* pParent = pNode->pParent;
	while(pParent && pParent->nCodeLength)
	{
		pNode->dwCode <<= 1;
		pNode->dwCode |= pParent->dwCode;
		pNode->nCodeLength++;
		pParent = pParent->pParent;
	}
}

int GetHuffmanTree(CHuffmanNode nodes[], bool bSetCodes = true)
{
	register CHuffmanNode* pNodes[256], *pNode;
	// add used ascii to Huffman queue
	register int nNodeCount = 0;
        register int nCount;
	for(nCount = 0; nCount < 256 && nodes[nCount].nFrequency; nCount++)
		pNodes[nNodeCount++] = &nodes[nCount];
	register int nParentNode = nNodeCount, nBackNode = nNodeCount-1;
	while(nBackNode > 0)
	{
		// parent node
		pNode = &nodes[nParentNode++];
		// pop first child
		pNode->pLeft = PopNode(pNodes, nBackNode--, false);
		// pop second child
		pNode->pRight = PopNode(pNodes, nBackNode--, true);
		// adjust parent of the two poped nodes
		pNode->pLeft->pParent = pNode->pRight->pParent = pNode;
		// adjust parent frequency
		pNode->nFrequency = pNode->pLeft->nFrequency + pNode->pRight->nFrequency;
		// insert parent node depending on its frequency
                register int i;
		for(i = nBackNode; i >= 0; i--)
			if(pNodes[i]->nFrequency >= pNode->nFrequency)
				break;
		memmove(pNodes+i+2, pNodes+i+1, (nBackNode-i)*sizeof(int));
		pNodes[i+1] = pNode;
		nBackNode++;
	}
	// set tree leaves nodes code
	if(bSetCodes)
		for(nCount = 0; nCount < nNodeCount; nCount++)
			SetNodeCode(&nodes[nCount]);

	return nNodeCount;
}

unsigned int CompressHuffman(unsigned char *pSrc, int nSrcLen, unsigned char **pDes)
{
        unsigned int nDesLen;

	CHuffmanNode nodes[511];
	// initialize nodes ascii
        int nCount;
	for(nCount = 0; nCount < 256; nCount++)
		nodes[nCount].byAscii = nCount;
	// get ascii frequencies
	for(nCount = 0; nCount < nSrcLen; nCount++)
		nodes[pSrc[nCount]].nFrequency++;
	// sort ascii chars depending on frequency
	qsort(nodes, 256, sizeof(CHuffmanNode), frequencyCompare);
	// construct Huffman tree
	int nNodeCount = GetHuffmanTree(nodes);
	// construct compressed buffer
	int nNodeSize = sizeof(unsigned int)+sizeof(unsigned char);
	nDesLen = nSrcLen+nNodeCount*nNodeSize;
	*pDes = (unsigned char*)malloc(nDesLen);
	unsigned char *pDesPtr = *pDes;
	memset(pDesPtr, 0, nDesLen);
	// save source buffer length at the first DWORD
	*(unsigned int*)pDesPtr = nSrcLen;
	pDesPtr += sizeof(unsigned int);
	// save Huffman tree leaves count-1 (as it may be 256)
	*pDesPtr = nNodeCount-1;
	pDesPtr += sizeof(unsigned char);
	// save Huffman tree used leaves nodes
	for(nCount = 0; nCount < nNodeCount; nCount++)
	{	// the array sorted on frequency so used nodes come first
		memcpy(pDesPtr, &nodes[nCount], nNodeSize);
		pDesPtr += nNodeSize;
	}
	// sort nodes depending on ascii to can index nodes with its ascii value
	qsort(nodes, 256, sizeof(CHuffmanNode), asciiCompare);

	int nDesIndex = 0;
	// loop to write codes
	for(nCount = 0; nCount < nSrcLen; nCount++)
	{
		*(unsigned int*)(pDesPtr+(nDesIndex>>3)) |= nodes[pSrc[nCount]].dwCode << (nDesIndex&7);
		nDesIndex += nodes[pSrc[nCount]].nCodeLength;
	}
	// update destination length
	nDesLen = (pDesPtr-*pDes)+(nDesIndex+7)/8;
	*pDes = (unsigned char*)realloc(*pDes,nDesLen);

        return nDesLen;
}

unsigned int DecompressHuffman(unsigned char *pSrc, int nSrcLen, unsigned char *pDes)
{
	// copy destination final length
        register unsigned int nDesLen=*(unsigned int*)pSrc;
	
	register int nNodeCount = *(pSrc+sizeof(unsigned int))+1;
	// initialize Huffman nodes with frequency and ascii
	register CHuffmanNode nodes[511], *pNode;
	register int nNodeSize = sizeof(unsigned int)+sizeof(unsigned char), nSrcIndex = nNodeSize;
	for(register int nCount = 0; nCount < nNodeCount; nCount++)
	{
		memcpy(&nodes[nCount], pSrc+nSrcIndex, nNodeSize);
		nSrcIndex += nNodeSize;
	}
	// construct Huffman tree
	GetHuffmanTree(nodes, false);
	// get Huffman tree root
	register CHuffmanNode *pRoot = &nodes[0];
	while(pRoot->pParent)
		pRoot = pRoot->pParent;

	register int nDesIndex = 0;
	register unsigned int nCode;
	nSrcIndex <<= 3;	// convert from bits to bytes
	while(nDesIndex < nDesLen)
	{
		nCode = (*(unsigned int*)(pSrc+(nSrcIndex>>3)))>>(nSrcIndex&7);
		pNode = pRoot;
		while(pNode->pLeft)	// if node has pLeft then it must has pRight
		{	// node not leaf
			pNode = (nCode&1) ? pNode->pRight : pNode->pLeft;
			nCode >>= 1;
			nSrcIndex++;
		}
		pDes[nDesIndex++] = pNode->byAscii;
	}

        return nDesLen;
}
